/*********************************************************************/
/* file: history.c - functions for the history stuff                 */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#endif
#ifdef TERMIOS
  #include <sys/termio.h>
#else
  #include <sys/ioctl.h>
#endif

#include "structs.h"
#include "history.h"
#include "parse.h"
#include "screen.h"
#include "utility.h"

extern struct master_table_data master_table;
extern int linepos;

/************************/
/* the #history command */
/************************/
session *history_command(const unsigned char *dummy, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  int i;

  if(HISTORY_SIZE - 1 > 24)
    i = 24;
  else
    i = HISTORY_SIZE - 1;

  for(; i >= 0; i--)
    if(master_table.history[i]) {
      sprintf((char *)buf, "%2d %s\n\r", i, master_table.history[i]);
      ses->print(buf);
    }

  return(ses);
}

void do_history(unsigned char *buffer, window *activ)
{
  unsigned char result[BUFFER_SIZE], *org;
  const unsigned char *cptr;
  int len;
  session *ses;
  
  ses = activ->get_active_ses();

  activ->set_ptexrem();
  activ->print((const uchar *)"\n\r");
  if(ses->get_output() != activ) {
    len = strlen((const char *)buffer);
    str_cat(buffer, len, BUFFER_SIZE, (const uchar *)"\n\r");
    ses->print(buffer);
    str_cat(buffer, len, BUFFER_SIZE, (const uchar *)"");
  } else {
    if(ses->term_echoing)
      ses->insert_ses_scrhist(buffer, TRUE);
  }

  if(*buffer) {
    org = buffer;
    cptr = space_out(buffer);
    
    if(*cptr) {
      if(*cptr == '!') {
	if(*(cptr+1) == '!') {
	  if(master_table.history[0]) {
	    strcpy((char *)result, (const char *)master_table.history[0]);
	    strcat((char *)result, (const char *)cptr+2);
	    strcpy((char *)org, (const char *)result);
	  }
	} else if(isdigit(*(cptr + 1))) {
	  int i = atoi((const char *)cptr + 1);
	  if(i >= 0 && i < HISTORY_SIZE && master_table.history[i]) {
	    strcpy((char *)result, (const char *)master_table.history[i]);
	    strcat((char *)result, (const char *)cptr+2);
	    strcpy((char *)org, (const char *)result);
	  }
	} else {
	  int i;
	  
	  for(i = 0; i < HISTORY_SIZE && master_table.history[i]; i++)
	    if(is_abrev(master_table.history[i], cptr + 1)) {
	      strcpy((char *)org, (const char *)master_table.history[i]);
	      master_table.history_pos = i + 1;
	      break;
	    }
	}
      }
      insert_history(org, ses);
    }
  }
}

/***********************************************/
/* insert buffer into a session`s history list */
/***********************************************/
void insert_history(unsigned char *buffer, struct session *ses)
{
  unsigned char *tmp;
  int i;

  if(master_table.history_pos &&
     master_table.history[master_table.history_pos - 1] &&
     strcmp((const char *)master_table.history[master_table.history_pos - 1],
	    (const char *)buffer) == 0) {

    tmp = master_table.history[master_table.history_pos - 1];
    for(i = master_table.history_pos - 1; i > 0; i--)
      master_table.history[i] = master_table.history[i - 1];

    master_table.history[0] = tmp;

  } else {
    if(master_table.history[HISTORY_SIZE -1])
      free(master_table.history[HISTORY_SIZE -1]);

    for(i = HISTORY_SIZE - 1; i > 0; i--)
      master_table.history[i] = master_table.history[i - 1];

    master_table.history[0] = (uchar *)strdup((const char *)buffer);
  }

  master_table.history_pos = 0;
}

// *****************************
// insert session screen history
// *****************************

void session::insert_ses_scrhist(const unsigned char *text, const int flu)
{
  unsigned char buf[BUFFER_SIZE], *str;
  int tmp, lines, len, cols, WinEnd;
  position orgpos;
  window *win;

  len = strip_ctrl_codes(buf, text);

  str = buf;
  cols = (MAX_LINE_LENGTH - 1 > window::get_cols() ?
	  window::get_cols() :
	  MAX_LINE_LENGTH - 1);
  while(*str || (flu && current_pos > 0)) {
    while(*str && *str != '\n' && current_pos < cols)
      current_line[current_pos++] = *str++;
    
    current_line[current_pos] = '\0';
    
    if(flu || *str == '\n' || current_pos == cols) {
      if(*str == '\n')
	str++;
      
      if(session_hist_num == MAX_SES_HIST) {
	if(history[session_hist_end]) {
	  free(history[session_hist_end]);
	  history[session_hist_end] = 0;
	}
      } else
	session_hist_num++;
      
      if(current_pos) {
	history[session_hist_end++] = (uchar *)strdup((const char *)current_line);
      } else {
	history[session_hist_end++] = (uchar *)strdup("");
      }
      current_line[0] = '\0';
      current_pos     = 0;
      
      if(session_hist_end == MAX_SES_HIST)
	session_hist_end = 0;
      
      session_hist_mark_pos++;
      if(session_hist_mark_pos >= MAX_SES_HIST) {
	session_hist_mark_pos = MAX_SES_HIST - 1;
	session_hist_mark_col = 1;
      }

      for(win = window::get_windows(); win; win = win->get_next()) {
	if(win->get_hist_on() == this) {
	  WinEnd = win->get_size()->end;
	  if(win->get_time_on())
	    WinEnd--;
	  
	  lines = WinEnd - win->get_size()->start;
	  
	  if(win->get_hist_pos() + lines < MAX_SES_HIST) {
	    win->add_hist_pos();
	  } else {
	    win->scroll_region(win->get_size()->start, WinEnd - 1);
	    
	    orgpos = *(win->get_cursor());
	    
	    if(window::get_scrcursor()->col != window::get_cols() ||
	       window::get_scrcursor()->row != WinEnd - 1)
	      win->move(WinEnd - 1, window::get_cols());
	    
	    tmp = session_hist_end - win->get_hist_pos();
	    if(tmp < 0)
	      tmp += MAX_SES_HIST;
	    
	    printf("\n\r%s", history[tmp]);
	    if(orgpos.row > win->get_size()->start)
	      orgpos.row--;
	    
	    win->move(&orgpos);
	  }
	  win->print_win_hist_line(FALSE);
	}
      }
    }
  }
}

const unsigned char *session::search_hist_word(const unsigned char *word, 
					       int &amount) const
{
  unsigned char buffer[MAX_LINE_LENGTH], newword[MAX_LINE_LENGTH], *str;
  const unsigned char *f_word = 0, *line, *word_start;
  int histpos, lines, maxlines, len, fullword = FALSE;
  
  amount = 0;
  maxlines = output->get_size()->end - output->get_size()->start + 1;
  
  len = strlen((const char *)word);
  histpos = session_hist_end;
  
  if(session_hist_num < maxlines)
    maxlines = session_hist_num;
  
  for(lines = 0; lines < maxlines; lines++) {
    histpos--;
    if(histpos < 0)
      histpos += MAX_SES_HIST;
    
    line = history[histpos];
    
    while(*line) {
      while(*line && !ownisalnum(*line))
	line++;
      
      word_start = line;
      str = newword;
      while(*line && ownisalnum(*line))
	*str++ = *line++;
      *str = '\0';
      
      if(is_abrev(newword, word)) {
	if(f_word) {
	  for(amount = len;
	      *(buffer + amount) && *(buffer + amount) == *(newword + amount);
	      amount++)
	    ;
	  
	  if(ownisalnum(*(f_word + amount)) || *(newword + amount))
	    fullword = FALSE;
	  *(buffer + amount) = '\0';
	} else {
	  amount = str_cat(buffer, 0, sizeof(buffer), newword);
	  f_word = word_start;
	  fullword = TRUE;
	}
      }
    }
  }
  if(fullword)
    amount++;
  return(f_word);
}
