/*********************************************************************/
/* file: misc.c - misc commands                                      */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifndef NSTDLIBS
#  include <stdlib.h>
#  include <unistd.h>
#endif

#include "structs.h"
#include "action.h"
#include "comms_help.h"
#include "llist.h"
#include "main.h"
#include "misc.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "utility.h"
#include "utils.h"


const uchar *letters = (const uchar *)" 0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]abcdefghijklmnopqrstuvwxyz{|}";
int letters_number = ']' - '0' + '}' - 'a' + 3;

/* externs */

extern struct master_table_data master_table;
extern unsigned char tintin_char;
extern unsigned char tintin_retchar;
extern unsigned char text_start, text_end;
extern unsigned char verbatim_char;
extern int speedwalk;
extern int default_chars_per_line;
extern int default_lines_in_scr;
extern struct sgttyb original_state;
extern unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,....$9 variables */
extern unsigned char *version_num;
extern struct command_elem *commands;

extern const uchar *CLEARSCREEN;
extern const uchar *MAXSCROLLSIZE;

/********************/
/* the #all command */
/********************/
session *all_command(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  session *sesptr;

  arg = get_arg_with_spaces(arg, buffer);
  for(sesptr = session::get_sessions(); sesptr; sesptr = sesptr->get_next()) {
    if(!parse_input(buffer, sesptr)) {
      myquitsig(0);
    }
  }

  return(ses);
}

/*********************/
/* the #bell command */
/*********************/
session *comm_bell(const unsigned char *dummy, session *ses)
{
  ses->print((const uchar *)"\a");
  return(ses);
}

/*********************/
/* the #if   command */
/*********************/
session *comm_if(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], condition[BUFFER_SIZE], thentext[BUFFER_SIZE], elsetext[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, condition);
  arg = get_arg_stop_spaces(arg, thentext);
  arg = get_arg_with_spaces(arg, elsetext);
  if(!*thentext) {
    ses->print((const uchar *)"#Error - Usage: #if <condition> <then text> [else text]\n\r");
    return(ses);
  }

  if(string_to_number(condition, ses)) {
    if(IS_SET(ses->echo_flags, ECHO_IF)) {
      sprintf((char *)buf, "[IF: %s]\n\r", thentext);
      ses->print(buf);
    }
    return(parse_input(thentext, ses));
  } else if(*elsetext) {
    if(IS_SET(ses->echo_flags, ECHO_IF)) {
      sprintf((char *)buf, "[IF ELSE: %s]\n\r", elsetext);
      ses->print(buf);
    }
    return(parse_input(elsetext, ses));
  }
  return(ses);
}

/***********************/
/* the #cmp    command */
/***********************/
session *comm_cmp(const unsigned char *arg, session *ses)
{
  unsigned char buf1[BUFFER_SIZE], buf2[BUFFER_SIZE], buf3[BUFFER_SIZE];
  unsigned char thentext[BUFFER_SIZE], elsetext[BUFFER_SIZE];
  const unsigned char *tmp;
  int icase = FALSE, groupvar = FALSE, result;

  arg = get_arg_stop_spaces(arg, buf1);
  if(*buf1 == '-') {
    tmp = buf1 + 1;
    while(isspace(*tmp))
      tmp++;
    
    while(*tmp) {
      switch(*tmp) {
      case 'i':
      case 'I':
	icase = TRUE;
	break;
      case 'g':
      case 'G':
	groupvar = TRUE;
	break;
      default:
	ses->print((const uchar *)"#Unknown flag in cmp command.\n\r");
	break;
      }
      tmp++;
      while(isspace(*tmp))
	tmp++;
    }
    
    arg = get_arg_stop_spaces(arg, buf1);
  }
  arg = get_arg_stop_spaces(arg, buf2);
  arg = get_arg_stop_spaces(arg, thentext);
  arg = get_arg_with_spaces(arg, elsetext);
  if(!*thentext) {
    ses->print(help_cmp);
    return(ses);
  }
  result = FALSE;
  if(groupvar) {
    tmp = buf2;
    if(icase) {
      do {
	tmp = get_arg_stop_spaces(tmp, buf3);
	result = (stricmp(buf1, buf3) == 0);
      } while(*tmp && !result);
    } else {
      do {
	tmp = get_arg_stop_spaces(tmp, buf3);
	result = (strcmp((const char *)buf1, (const char *)buf3) == 0);
      } while(*tmp && !result);
    }
  } else {
    if(icase)
      result = (stricmp(buf1, buf2) == 0);
    else
      result = (strcmp((const char *)buf1, (const char *)buf2) == 0);
  }
  
  if(result) {
    if(IS_SET(ses->echo_flags, ECHO_CMP)) {
      sprintf((char *)buf1, "[CMP: %s]\n\r", thentext);
      ses->print(buf1);
    }
    return(parse_input(thentext, ses));
  } else if(*elsetext) {
    if(IS_SET(ses->echo_flags, ECHO_CMP)) {
      sprintf((char *)buf1, "[CMP ELSE: %s]\n\r", elsetext);
      ses->print(buf1);
    }
    return(parse_input(elsetext, ses));
  }
  return(ses);
}

session *comm_for(const unsigned char *arg, session *ses)
{
  unsigned char buf1[BUFFER_SIZE], buf2[BUFFER_SIZE], text[BUFFER_SIZE];
  int num1, num2, i;

  arg = get_arg_stop_spaces(arg, buf1);
  arg = get_arg_stop_spaces(arg, buf2);
  arg = get_arg_with_spaces(arg, text);

  if(!*text) {
    ses->print((const uchar *)"#Error - Usage: #for <number> <number> <text>\n\r");
    ses->print((const uchar *)"#        Use $0 mark number position(s) in text\n\r");
    return(ses);
  }
  num1 = string_to_number(buf1, ses);
  num2 = string_to_number(buf2, ses);

  for(i = num1; i <= num2; i++) {
    sprintf((char *)vars[0], "%d", i);
    substitute_vars(text, buf1, sizeof(buf1), TRUE, ses, FALSE);
    if(!(ses = parse_input(buf1, ses)))
      break;
  }
  return(ses);
}

/* ******************** * the clear command * ******************** */
session *clear_command(const unsigned char *dummy, session *ses)
{
  ses->get_output()->clear();
  ses->get_output()->print_prompt(TRUE, FALSE);
  return(ses);
}

/*********************/
/* the #boss command */
/*********************/
session *boss_command(const unsigned char *dummy, session *ses)
{
  char letter;
  int i;

  printf("%s%s", MAXSCROLLSIZE, CLEARSCREEN);
  for(i = 0; i < window::get_rows(); i++)
    printf("in-order traverse of tree starting from node %d resulted in %d red nodes\n\r", i, window::get_rows() - i);
  
  /* stop screen from scrolling stuff */
  do {
    letter = getchar();
  } while(letter != '\n' && letter != '\r' && letter != ' ');
  
  redraw_screen();
  return(ses);
}

/*********************/
/* the #char command */
/*********************/
session *char_command(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    tintin_char = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW TINTIN-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER TINTIN-CHAR! SOMETHING LIKE # OR /!\n\r");

  return(ses);
}

/*********************/
/* the #text command */
/*********************/
session *comm_text(const unsigned char *arg, session *ses)
{
  arg = space_out(arg);
  if(ispunct(*arg) && ispunct(*(arg + 1))) {
    if(*arg != *(arg + 1)) {
      text_start = *arg++;
      text_end =*arg;
      ses->print((const uchar *)"#OK. GOT NEW START AND END CHAR.\n\r");
    } else
      ses->print((const uchar *)"#Start and end char must be diffirent!\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER CHARS! SOMETHING LIKE () OR []!\n\r");

  return(ses);
}

/***************************/
/* the #returnchar command */
/***************************/
session *comm_returnchar(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    tintin_retchar = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW TINTIN-RETURN-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER TINTIN-RETURN-CHAR! SOMETHING LIKE ; OR :\n\r");
  return(ses);
}

/*********************/
/* the #echo command */
/*********************/
session *echo_command(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp;
  unsigned char *s_ac = (uchar *)"ACTION";
  unsigned char *s_i  = (uchar *)"IF";
  unsigned char *s_cm = (uchar *)"CMP";
  unsigned char *s_w  = (uchar *)"WAIT";
  unsigned char *s_al = (uchar *)"ALIAS";
  int *echo, flag, all = FALSE;

  echo = &ses->echo_flags;

  arg = get_arg_stop_spaces(arg, buf);

  switch(*buf) {
  case 'a': {
    switch(*(buf + 1)) {
    case 'c': {
      flag = ECHO_ACTION;
      tmp = s_ac;
    } break;
    case 'l': {
      if(*(buf + 2) == 'l' && !*(buf + 3)) {
	flag = 0;
	tmp = 0;
	all = TRUE;
      } else {
	flag = ECHO_ALIAS;
	tmp = s_al;
      }
    } break;
    default: {
      ses->print(help_echo);
      return(ses);
    } break;
    } /* switch */
  } break;
  case 'i': {
    flag = ECHO_IF;
    tmp = s_i;
  } break;
  case 'c': {
    flag = ECHO_CMP;
    tmp = s_cm;
  } break;
  case 'w': {
    flag = ECHO_WAIT;
    tmp = s_w;
  } break;
  default: {
    ses->print(help_echo);
    return(ses);
  } break;
  } /* switch */

  arg = get_arg_with_spaces(arg, buf);
  if(*buf || all) {
    if(is_abrev((const uchar *)"on", buf) ||
       is_abrev((const uchar *)"yes", buf)) {
      if(all) {
	SET_BIT(*echo, ECHO_ACTION | ECHO_ALIAS | ECHO_WAIT | ECHO_IF | ECHO_CMP);
	sprintf((char *)buf, "#OK. ALL ECHO IS NOW ON.\n\r");
      } else {
	SET_BIT(*echo, flag);
	sprintf((char *)buf, "#OK. %s ECHO IS NOW ON.\n\r", tmp);
      }
    } else if(is_abrev((const uchar *)"off", buf) ||
	      is_abrev((const uchar *)"no", buf)) {
      if(all) {
	*echo = 0;
	sprintf((char *)buf, "#OK. ALL ECHO IS NOW OFF.\n\r");
      } else {
	REMOVE_BIT(*echo, flag);
	sprintf((char *)buf, "#OK. %s ECHO IS NOW OFF.\n\r", tmp);
      }
    } else {
      sprintf((char *)buf, "#Use 'on', 'yes', 'off' or 'no'\n\r");
    }
  } else {
    if(!IS_SET(*echo, flag)) {
      SET_BIT(*echo, flag);
      sprintf((char *)buf, "#OK. %s ECHO IS NOW ON.\n\r", tmp);
    } else {
      REMOVE_BIT(*echo, flag);
      sprintf((char *)buf, "#OK. %s ECHO IS NOW OFF.\n\r", tmp);
    }
  }
  ses->print(buf);
  return(ses);
}

// ********************
// the #message command
// ********************
session *comm_message(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp;
  unsigned char *s_ac = (uchar *)"ACTION";
  unsigned char *s_s  = (uchar *)"SUBSTITUTE";
  unsigned char *s_w  = (uchar *)"WAIT";
  unsigned char *s_v  = (uchar *)"VARIABLE";
  unsigned char *s_al = (uchar *)"ALIAS";
  int *message, flag, all = FALSE;

  message = &ses->message_flags;

  arg = get_arg_stop_spaces(arg, buf);

  switch(*buf) {
  case 'a': {
    switch(*(buf + 1)) {
    case 'c': {
      flag = MESSAGE_ACTION;
      tmp = s_ac;
    } break;
    case 'l': {
      if(*(buf + 2) == 'l' && !*(buf + 3)) {
	flag = 0;
	tmp = 0;
	all = TRUE;
      } else {
	flag = MESSAGE_ALIAS;
	tmp = s_al;
      }
    } break;
    default: {
      ses->print(help_message);
      return(ses);
    } break;
    } /* switch */
  } break;
  case 's': {
    flag = MESSAGE_SUBS;
    tmp = s_s;
  } break;
  case 'w': {
    flag = MESSAGE_WAIT;
    tmp = s_w;
  } break;
  case 'v': {
    flag = MESSAGE_VAR;
    tmp = s_v;
  } break;
  default: {
    ses->print(help_message);
    return(ses);
  } break;
  } /* switch */

  arg = get_arg_with_spaces(arg, buf);
  if(*buf || all) {
    if(is_abrev((const uchar *)"on", buf) ||
       is_abrev((const uchar *)"yes", buf)) {
      if(all) {
	SET_BIT(*message, MESSAGE_ACTION | MESSAGE_ALIAS | MESSAGE_SUBS | MESSAGE_WAIT | MESSAGE_VAR);
	sprintf((char *)buf, "#OK. ALL DEFINE MESSAGES ARE NOW ON.\n\r");
      } else {
	SET_BIT(*message, flag);
	sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW ON.\n\r", tmp);
      }
    } else if(is_abrev((const uchar *)"off", buf) ||
	      is_abrev((const uchar *)"no", buf)) {
      if(all) {
	*message = 0;
	sprintf((char *)buf, "#OK. ALL DEFINE MESSAGES ARE NOW OFF.\n\r");
      } else {
	REMOVE_BIT(*message, flag);
	sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW OFF.\n\r", tmp);
      }
    } else {
      sprintf((char *)buf, "#Use 'on', 'yes', 'off' or 'no'\n\r");
    }
  } else {
    if(!IS_SET(*message, flag)) {
      SET_BIT(*message, flag);
      sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW ON.\n\r", tmp);
    } else {
      REMOVE_BIT(*message, flag);
      sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW OFF.\n\r", tmp);
    }
  }
  ses->print(buf);
  return(ses);
}

/*********************/
/* the #end command */
/*********************/
session *end_command(const unsigned char *arg, session *ses)
{
  exit_tintin(0);
  return(0);
}

/***********************/
/* the #ignore command */
/***********************/
session *ignore_command(const unsigned char *arg,session *ses)
{
  unsigned char buf[100], *tmp;
  unsigned char s_ac[10] = "ACTION";
  unsigned char s_su[10] = "SUBS";
  unsigned char s_al[10] = "ALIAS";
  unsigned char s_va[10] = "VARIABLE";
  unsigned char s_ca[10] = "CALCULATE";
  int *ignore, flag;

  ignore = &ses->ignore_flags;

  arg = space_out(arg);

  switch(*arg) {
  case 'a': {
    switch(*(arg + 1)) {
    case 'c': {
      flag = IGNORE_ACTION;
      tmp = s_ac;
    } break;
    case 'l': {
      flag = IGNORE_ALIAS;
      tmp = s_al;
    } break;
    default: {
      ses->print(help_ignore);
      return(ses);
    } break;
    } /* switch */
  } break;
  case 'c': {
    flag = IGNORE_CALC;
    tmp = s_ca;
  } break;
  case 's': {
    flag = IGNORE_SUBS;
    tmp = s_su;
  } break;
  case 'v': {
    flag = IGNORE_VAR;
    tmp = s_va;
  } break;
  default: {
    ses->print(help_ignore);
    return(ses);
  } break;
  } /* switch */

  if(!IS_SET(*ignore, flag)) {
    SET_BIT(*ignore, flag);
    sprintf((char *)buf, "#OK. %s IGNORE IS NOW ON.\n\r", tmp);
  } else {
    REMOVE_BIT(*ignore, flag);
    sprintf((char *)buf, "#OK. %s IGNORE IS NOW OFF.\n\r", tmp);
  }
  ses->print(buf);

  return(ses);
}

/**********************/
/* the #snoop command */
/**********************/
session *snoop_command(const unsigned char *arg, session *ses)
{
  unsigned char buf[100], buffer[BUFFER_SIZE];
  session *sesptr=ses;

  arg = get_arg_stop_spaces(arg, buffer);
  if(*buffer) {
    for(sesptr = session::get_sessions();
	sesptr && strcmp((const char *)sesptr->get_name(), (const char *)buffer);
	sesptr = sesptr->get_next());
    if(!sesptr) {
      ses->print((const uchar *)"#NO SESSION WITH THAT NAME!\n\r");
      return(ses);
    }
  }
  if(sesptr->snoopstatus) {
    sesptr->snoopstatus=FALSE;
    sprintf((char *)buf, "#UNSNOOPING SESSION '%s'\n\r", sesptr->get_name());
    ses->print(buf);
  } else {
    sesptr->snoopstatus=TRUE;
    sprintf((char *)buf, "#SNOOPING SESSION '%s'\n\r", sesptr->get_name());
    ses->print(buf);
  }

  return(ses);
}

/**************************/
/* the #speedwalk command */
/**************************/
session *speedwalk_command(const unsigned char *dummy, session *ses)
{
  speedwalk=!speedwalk;
  if(speedwalk)
    ses->print((const uchar *)"#SPEEDWALK IS NOW ON.\n\r");
  else
    ses->print((const uchar *)"#SPEEDWALK IS NOW OFF.\n\r");

  return(ses);
}

/***********************/
/* the #system command */
/***********************/
session *system_command(const unsigned char *arg, session *ses)
{
  char letter;
  unsigned char buffer[BUFFER_SIZE];
  
  sigmask_on();
  
  arg = get_arg_with_spaces(arg, buffer);
  if(*buffer) {
    ses->print((const uchar *)"#OK EXECUTING SHELL COMMAND.\n\r");
    ses->print((const uchar *)"#Press return after it to continue.\n\r");
    
    nocbreak_echo();
    system((const char *)buffer);
    
    do {
      letter = getchar();
    } while(letter != '\n' && letter != '\r');
    
    cbreak_noecho();
    
    ses->print((const uchar *)"#OK COMMAND EXECUTED.\n\r");
    get_screen_size();
  } else {

    if(!getenv(DEF_SHELL_ENV)) {
      ses->print((const uchar *)"#ERROR - COULDN'T GET DEFAULT SHELL NAME.\n\r");
    } else {
      
      ses->print((const uchar *)"#OK OPENING DEFAULT SHELL.\n\r");
      nocbreak_echo();
      system(getenv(DEF_SHELL_ENV));
      cbreak_noecho();
      
      ses->print((const uchar *)"#OK SHELL CLOSED.\n\r");
      get_screen_size();
    }
  }
  
  sigmask_off();
  return(ses);
}

/********************/
/* the #zap command */
/********************/
session *zap_command(const unsigned char *dummy, session *ses)
{
  window *win;

  win = ses->get_input();
  ses->print((const uchar *)"#ZZZZZZZAAAAAAAAPPPP!!!!!!!!! LET'S GET OUTTA HERE!!!!!!!!\n\r");
  delete ses;
  return(win->get_active_ses());
}

/*********************/
/* the #help command */
/*********************/

session *help_command(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[1024], buffer[BUFFER_SIZE], *str;
  int pos, comm_no;

  arg = get_arg_stop_spaces(arg, buffer);
  str = buffer;
  if(!*str) {
    sprintf((char *)tmp, "More help by typing: %chelp <COMMAND NAME>\n\r",
	    tintin_char);
    pos = str_cat(buf, 0, BUFFER_SIZE, tmp);
    pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"Possible commands: ");
    for(comm_no = 0; *commands[comm_no].name != '\n'; comm_no++) {
      if(comm_no % 5 == 0)
	pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"\n\r  ");
      sprintf((char *)tmp, "%c%c%-13.13s",
	      commands[comm_no].help_txt ? ' ' : '?',
	      tintin_char,
	      commands[comm_no].name);
      pos = str_cat(buf, pos, BUFFER_SIZE, tmp);
    }
    pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"\n\r");
    ses->print(buf);
  } else {
    do {
      if(*str == tintin_char)
	str++;
      
      for(comm_no = 0; *commands[comm_no].name != '\n'; comm_no++) {
	if(is_abrev(commands[comm_no].name, str)) {
	  if(commands[comm_no].help_txt) {
	    ses->print(commands[comm_no].help_txt);
	  } else {
	    ses->print((const uchar *)"#Sorry no help available for that command.\n\r");
	  }
	  break;
	}
      }
      if(*commands[comm_no].name == '\n') {
	sprintf((char *)tmp, "#There isn't such command as %c%s\n\r", tintin_char, str);
	ses->print(tmp);
      }
      arg = get_arg_stop_spaces(arg, buffer);
      str = buffer;
    } while(*str);
  }
  return(ses);
}

session *comm_resetscreen(const unsigned char *dummy, session *ses)
{
  cbreak_noecho();
  get_screen_size();
  ses->print((const uchar *)"OK.\n\r");
  return(ses);
}

/***********************/
/* the #showme command */
/***********************/
session *comm_showme(const unsigned char *arg, session *ses)
{
  unsigned char tmp[BUFFER_SIZE], result[BUFFER_SIZE], *str;
  int pos, check_act = FALSE;

  arg = get_arg_stop_spaces(arg, tmp);
  if(*tmp == '-') {
    str = tmp + 1;
    while(isspace(*str))
      str++;
    
    while(*str) {
      switch(*str) {
      case 'c':
      case 'C':
	check_act = TRUE;
	break;
      default:
	sprintf((char *)result, "#Unknown argument (%c).\n\r", *str);
	ses->print(result);
      }
      str++;
      
      while(isspace(*str))
	str++;
    }
    arg = get_arg_with_spaces(arg, tmp);
  }
  
//  pos = substitute_vars(tmp, result, sizeof(result), TRUE, ses, TRUE);
  
  if(check_act) {
    if(!(ses = process_text(tmp, result, pos, ses)))
      return(0);
    
    if(*result || !*tmp)
      pos = str_cat(result, pos, sizeof(result), (const uchar *)"\n\r");
  } else {
    pos = str_cat(result, 0, sizeof(result), tmp);
    pos = str_cat(result, pos, sizeof(result), (const uchar *)"\n\r");
  }
  ses->print(result);
  
  return(ses);
}

/**********************************
 * the #word and #unword commands *
 ******************************** */

session *comm_word(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[BUFFER_SIZE];
  int i, pos;
  
  while(isspace(*arg))
    arg++;
  
  while(*arg) {
    arg = get_arg_stop_spaces(arg, buf);
    
    if(*buf) {
      if(ses->word_list[WORD_LIST_SIZE - 1]) {
	ses->print((const uchar *)"#Word list already full.\n\r");
	break;
      }
      for(i = 0;
	  ses->word_list[i] &&
	  strcmp((const char *)buf, (const char *)ses->word_list[i]) > 0; i++)
	;
      if(ses->word_list[i] &&
	 strcmp((const char *)buf, (const char *)ses->word_list[i]) == 0) {
	sprintf((char *)tmp, "#Word '%s' is already in list\n\r", buf);
	ses->print(tmp);
      } else {
	sprintf((char *)tmp, "#Word '%s' added to list.\n\r", buf);
	ses->print(tmp);
	
	for(pos = i, i = WORD_LIST_SIZE - 1; i > pos; i--)
	  ses->word_list[i] = ses->word_list[i - 1];
	
	ses->word_list[pos] = (uchar *)strdup((const char *)buf);
      }
    }
    while(isspace(*arg))
      arg++;
  }
  
  ses->print((const uchar *)"#Ok.\n\r");
  return(ses);
}

session *comm_unword(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[BUFFER_SIZE];
  int pos, found, i;
  
  while(isspace(*arg))
    arg++;
  
  while(*arg) {
    arg = get_arg_stop_spaces(arg, buf);
    
    pos = 0;
    found = FALSE;
    while(pos < WORD_LIST_SIZE && ses->word_list[pos]) {
      if(mystrcmp(ses->word_list[pos], buf)) {
	sprintf((char *)tmp,
		"#Word '%s' removed from list.\n\r", ses->word_list[pos]);
	ses->print(tmp);
	
	free(ses->word_list[pos]);
	
	for(i = pos; i < WORD_LIST_SIZE - 1; i++)
	  ses->word_list[i] = ses->word_list[i + 1];
	ses->word_list[i] = 0;
	
	found = TRUE;
      } else
	pos++;
    }
    if(!found) {
      sprintf((char *)tmp, "#No '%s' word on list\n\r", buf);
      ses->print(tmp);
    }
    while(isspace(*arg))
      arg++;
  }
  ses->print((const uchar *)"#Ok.\n\r");
  return(ses);
}

/************************
 * the #version command *
 ********************** */

session *comm_version(const unsigned char *dummy, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  
  sprintf((char *)buf,
	  "********************************************************************\n\r"
	  " ht (HUT TinTin) Version %s\n\r"
	  "\n\r"
	  " If you found bugs or have new ideas what to add write:\n\r"
	  " Email: otulonen@snakemail.hut.fi\n\r"
	  "********************************************************************\n\r",
	  version_num);
  ses->print(buf);
  return(ses);
}

/************************************
 * the #crypt and #decrypt commands *
 ********************************** */

session *comm_crypt(const unsigned char *arg, session *ses)
{
  unsigned char command[BUFFER_SIZE + 20], text[BUFFER_SIZE];
  unsigned char cryptword[BUFFER_SIZE], word[BUFFER_SIZE], letter, *crword = 0;
  int i, j, pos;
  struct listnode *ln;
  
  if(ses->get_socket() >= 0) {
    arg = get_arg_stop_spaces(arg, command);
    arg = get_arg_stop_spaces(arg, text);
    arg = get_arg_with_spaces(arg, cryptword);
    
    if(!*text) {
      ses->print((const uchar *)"#No text to crypt.\n\r");
      return(ses);
    }
    
    if(*cryptword) {
      crword = cryptword;
    } else {
      if(!(ln = searchnode_list(ses->variables, (unsigned char *)"HUTCRYPT"))||
	 !*ln->right) {
	ses->print((const uchar *)"#You need to define 'HUTCRYPT' variable first.\n\r");
	return(ses);
      }
      crword = ln->right;
    }
    for(i = 0, j = 0; *(crword + i); i++) {
      letter = *(crword + i);
      if(letter == ' ')
	word[j++] = letter - ' ';
      else if(letter >= '0' && letter <= ']')
	word[j++] = letter - '0' + 1;
      else if(letter >= 'a' && letter <= '}')
	word[j++] = letter - 'a' + ']' - '0' + 2;
    }
    word[j] = '\0';
    
    if(j > 0) {
      pos = str_cat(command, strlen((const char *)command), sizeof(command), (const uchar *)" ~~^");
      
      for(i = 0, j = 0; text[i] != '\0'; i++) {
	letter = text[i];
	if(letter == ' ')
	  letter = letter - ' ';
	else if(letter >= '0' && letter <= ']')
	  letter = letter - '0' + 1;
	else if(letter >= 'a' && letter <= '}')
	  letter = letter - 'a' + ']' - '0' + 2;
	else {
	  if(letter != '~')
	    command[pos++] = letter;
	  continue;
	}
	command[pos++] = letters[(letter + word[j++]) % letters_number];
	if(word[j] == '\0')
	  j = 0;
      }
      pos = str_cat(command, pos, sizeof(command), (const uchar *)"^~~[");
      pos = str_cat(command, pos, sizeof(command), version_num);
      pos = str_cat(command, pos, sizeof(command), (const uchar *)"]");
      ses = ses->write_net(command);
    } else {
      ses->print((const uchar *)"#You have bad 'HUTCRYPT' word define it again\n\r");
    }
  } else {
    ses->print((const uchar *)"#There isn't active session.\n\r");
  }
  return(ses);
}

session *comm_decrypt(const unsigned char *arg, session *ses)
{
  unsigned char text[BUFFER_SIZE], word[BUFFER_SIZE], letter;
  unsigned char cryptword[BUFFER_SIZE], *crword = 0;
  int i, j;
  struct listnode *ln;
  
  arg = get_arg_stop_spaces(arg, text);
  arg = get_arg_with_spaces(arg, cryptword);
  
  if(!*text) {
    ses->print((const uchar *)"#No text to decrypt.\n\r");
    return(ses);
  }
  
  if(*cryptword) {
    crword = cryptword;
  } else {
    if(!(ln = searchnode_list(ses->variables, (unsigned char *)"HUTCRYPT")) ||
       !*ln->right) {
      ses->print((const uchar *)"#You need to define 'HUTCRYPT' variable first.\n\r");
      return(ses);
    }
    crword = ln->right;
  }
  for(i = 0, j = 0; *(crword + i); i++) {
    letter = *(crword + i);
    if(letter == ' ')
      word[j++] = letter - ' ';
    else if(letter >= '0' && letter <= ']')
      word[j++] = letter - '0' + 1;
    else if(letter >= 'a' && letter <= '}')
      word[j++] = letter - 'a' + ']' - '0' + 2;
  }
  word[j] = '\0';
  
  if(j > 0) {
    i = 0;
    j = 0;
    while(*(text + i)) {
      if(*(text + i) == '\\') {
	i++;
	if(*(text + i))
	  *(text + j++) = *(text + i++);
      } else
	*(text + j++) = *(text + i++);
    }
    *(text + j) = '\0';
    
    for(i = 0, j = 0; *(text + i); i++) {
      letter = *(text + i);
      
      if(letter == ' ')
	letter = letter - ' ';
      else if(letter >= '0' && letter <= ']')
	letter = letter - '0' + 1;
      else if(letter >= 'a' && letter <= '}')
	letter = letter - 'a' + ']' - '0' + 2;
      else {
	continue;
      }	
      if(letter - word[j] < 0) {
	*(text + i) = letters[letter - word[j] + letters_number];
      } else {
	*(text + i) = letters[letter - word[j]];
      }
      if(word[++j] == '\0')
	j = 0;
    }
    *(text + i) = '\0';
    sprintf((char *)word, "#Decrypted '%s'\n\r", text);
    ses->print(word);
  } else {
    ses->print((const uchar *)"#You have bad 'HUTCRYPT' word, use diffirent one\n\r");
  }
  return(ses);
}

// ****************************
// verbatim char change command
// ****************************

session *comm_verbatim(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  
  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    verbatim_char = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW VERBATIM-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER VERBATIM-CHAR! SOMETHING LIKE / OR \\!\n\r");
  
  return(ses);
}

// *************************
// the #verbatimmode command
// *************************
session *comm_verbatimmode(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    if(is_abrev((const uchar *)"on", buf) ||
       is_abrev((const uchar *)"yes", buf)) {
      master_table.verbatimmode = TRUE;
      ses->print((const uchar *)"#Verbatimmode on.\n\r");
    } else if(is_abrev((const uchar *)"off", buf) ||
	      is_abrev((const uchar *)"no", buf)) {
      master_table.verbatimmode = FALSE;
      ses->print((const uchar *)"#Verbatimmode off.\n\r");
    }
  } else {
    if(!master_table.verbatimmode) {
      master_table.verbatimmode = TRUE;
      ses->print((const uchar *)"#Verbatimmode on.\n\r");
    } else {
      master_table.verbatimmode = FALSE;
      ses->print((const uchar *)"#Verbatimmode off.\n\r");
    }
  }
  return(ses);
}

// ************************
//  the alive command code
// ************************

session *comm_alive(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  struct listnode **myvars, **mycolors, *ln;
  session *sesptr;
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    myvars = &ses->variables;
    mycolors = &ses->colors;
    
    if((ln = searchnode_list(*mycolors, (unsigned char*)"STATUS")) != NULL)
      delete ln;
    
    if((ln = searchnode_list(*myvars, (unsigned char*)"STATUS")) != NULL)
      delete ln;
    
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)buf) == 0)
	break;
    
    if(sesptr)
      ln = new listnode(myvars, (unsigned char*)"STATUS", (unsigned char*)"1",
			TYPE_VAR, 5);
    else
      ln = new listnode(myvars, (unsigned char*)"STATUS", (unsigned char*)"0",
			TYPE_VAR, 5);
    
    if(IS_SET(ses->message_flags, MESSAGE_VAR)) {
      ses->print((const uchar *)"#OK. STATUS VARIABLE SET.\n\r");
    }
  } else {
    ses->print(help_alive);
  }
  return(ses);
}

// ********************
//  no_neg_off command
// ********************

session *comm_no_neg_off(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    if(is_abrev((const uchar *)"on", buf) ||
       is_abrev((const uchar *)"yes", buf)) {
      master_table.no_neg_off = TRUE;
      ses->print((const uchar *)"#No negative off mode on.\n\r");
    } else if(is_abrev((const uchar *)"off", buf) ||
	      is_abrev((const uchar *)"no", buf)) {
      master_table.no_neg_off = FALSE;
      ses->print((const uchar *)"#No negative off mode off.\n\r");
    }
  } else {
    if(!master_table.no_neg_off) {
      master_table.no_neg_off = TRUE;
      ses->print((const uchar *)"#No negative off mode on.\n\r");
    } else {
      master_table.no_neg_off = FALSE;
      ses->print((const uchar *)"#No negative off mode off.\n\r");
    }
  }
  return(ses);
}
