/*********************************************************************/
/* file: parse.c - some utility-functions                            */
/*                             TINTIN III                            */
/*          (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t             */
/*                     coded by peter unold 1992                     */
/*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#endif
#ifdef OSF1v3
typedef int clockid_t;
#endif
#include <time.h>
#include <assert.h>

#include "structs.h"
#include "action.h"
#include "alias.h"
#include "color.h"
#include "comms_help.h"
#include "files.h"
#include "history.h"
#include "llist.h"
#include "main.h"
#include "misc.h"
#include "net.h"
#include "parse.h"
#include "path.h"
#include "screen.h"
#include "session.h"
#include "substitute.h"
#include "ticks.h"
#include "utility.h"
#include "utils.h"
#include "variable.h"

extern struct master_table_data master_table;
extern unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,....$9 variables */
extern unsigned char tintin_char;
extern unsigned char tintin_retchar;
extern unsigned char text_start, text_end;
extern unsigned char verbatim_char;
extern int speedwalk;

const unsigned char *comm_names[] = {
  (const uchar *)"action",
  (const uchar *)"alias",
  (const uchar *)"alive",
  (const uchar *)"all",
  (const uchar *)"autolog",
  (const uchar *)"bell",
  (const uchar *)"boss",
  (const uchar *)"clear",
  (const uchar *)"char",
  (const uchar *)"close",
  (const uchar *)"cmp",
  (const uchar *)"color",
  (const uchar *)"crypt",
  (const uchar *)"ctrl",
  (const uchar *)"decrypt",
  (const uchar *)"echo",
  (const uchar *)"en",
  (const uchar *)"end",
  (const uchar *)"for",
  (const uchar *)"help",
  (const uchar *)"history",
  (const uchar *)"if",
  (const uchar *)"ignore",
  (const uchar *)"input",
  (const uchar *)"log",
  (const uchar *)"logline",
  (const uchar *)"logseshist",
  (const uchar *)"mark",
  (const uchar *)"message",
  (const uchar *)"nop",
  (const uchar *)"no_neg_off",
  (const uchar *)"output",
  (const uchar *)"path",
  (const uchar *)"return",
  (const uchar *)"read",
  (const uchar *)"repeatif",
  (const uchar *)"resetscreen",
  (const uchar *)"returnchar",
  (const uchar *)"savepath",
  (const uchar *)"session",
  (const uchar *)"showme",
  (const uchar *)"snoop",
  (const uchar *)"speedwalk",
  (const uchar *)"split",
  (const uchar *)"substitute",
  (const uchar *)"text",
  (const uchar *)"tick",
  (const uchar *)"tickoff",
  (const uchar *)"tickon",
  (const uchar *)"tickset",
  (const uchar *)"ticksize",
  (const uchar *)"time",
  (const uchar *)"unaction",
  (const uchar *)"unalias",
  (const uchar *)"unpath",
  (const uchar *)"unsubstitute",
  (const uchar *)"unvariable",
  (const uchar *)"unword",
  (const uchar *)"variable",
  (const uchar *)"verbatim",
  (const uchar *)"verbatimmode",
  (const uchar *)"version",
  (const uchar *)"wait",
  (const uchar *)"wininfo",
  (const uchar *)"winsize",
  (const uchar *)"word",
  (const uchar *)"write",
  (const uchar *)"writesession",
  (const uchar *)"zap",
  (const uchar *)"\n"
  };

struct command_elem *commands;

// ***************
//   nop command
// ***************

session *comm_nop(const unsigned char *dummy, session *ses)
{
  return(ses);
}

// *****************************
//       repeatif command
// real code is in parse_input()
// *****************************

session *comm_repeatif(const unsigned char *dummy, session *ses)
{
  assert(FALSE);  // should NEVER get here...
  return(ses);
}

void init_comms_table(void)
{
  int i, pos;

#define COMM(help, cf) { commands[pos].name = comm_names[pos]; \
			   commands[pos].help_txt = (help); \
			     commands[pos++].func = (cf); }

  for(i = 0; *comm_names[i] != '\n'; i++)
    ;
  commands = new command_elem[i + 1];

  pos = 0;
  COMM(help_action,       action_command);
  COMM(help_alias,        alias_command);
  COMM(help_alive,        comm_alive);
  COMM(help_all,          all_command);
  COMM(help_autolog,      comm_autolog);
  COMM(help_bell,         comm_bell);
  COMM(help_boss,         boss_command);
  COMM(help_clear,        clear_command);
  COMM(help_char,         char_command);
  COMM(help_close,        comm_close);
  COMM(help_cmp,          comm_cmp);
  COMM(help_color,        comm_color); 
  COMM(help_crypt,        comm_crypt);
  COMM(help_ctrl,         comm_ctrl);
  COMM(help_decrypt,      comm_decrypt);
  COMM(help_echo,         echo_command);
  COMM(0,                 0);
  COMM(help_end,          end_command);
  COMM(help_for,          comm_for);
  COMM(help_help,         help_command);
  COMM(help_history,      history_command);
  COMM(help_if,           comm_if);
  COMM(help_ignore,       ignore_command);
  COMM(help_input,        comm_input);
  COMM(help_log,          log_command);
  COMM(help_logline,      comm_logline);
  COMM(help_logseshist,   comm_logseshist);
  COMM(help_mark,         mark_command);
  COMM(help_message,      comm_message);
  COMM(help_nop,          comm_nop);
  COMM(help_no_neg_off,   comm_no_neg_off);
  COMM(help_output,       comm_output);
  COMM(help_path,         path_command);
  COMM(help_return,       return_command);
  COMM(help_read,         read_command);
  COMM(help_repeatif,     comm_repeatif);
  COMM(help_resetscreen,  comm_resetscreen);
  COMM(help_returnchar,   comm_returnchar);
  COMM(help_savepath,     comm_savepath);
  COMM(help_session,      session_command);
  COMM(help_showme,       comm_showme);
  COMM(help_snoop,        snoop_command);
  COMM(help_speedwalk,    speedwalk_command);
  COMM(help_split,        comm_split);
  COMM(help_substitute,   parse_sub);
  COMM(help_text,         comm_text);
  COMM(help_tick,         tick_command);
  COMM(help_tickoff,      tickoff_command);
  COMM(help_tickon,       tickon_command);
  COMM(help_tickset,      tickset_command);
  COMM(help_ticksize,     ticksize_command);
  COMM(help_time,         comm_time);
  COMM(help_unaction,     unaction_command);
  COMM(help_unalias,      unalias_command);
  COMM(help_unpath,       unpath_command);
  COMM(help_unsubstitute, unsubstitute_command);
  COMM(help_unvariable,   comm_unvariable);
  COMM(help_unword,       comm_unword);
  COMM(help_variable,     comm_variable);
  COMM(help_verbatim,     comm_verbatim);
  COMM(help_verbatimmode, comm_verbatimmode);
  COMM(help_version,      comm_version);
  COMM(help_wait,         comm_wait);
  COMM(help_wininfo,      comm_wininfo);
  COMM(help_winsize,      comm_winsize);
  COMM(help_word,         comm_word);
  COMM(help_write,        write_command);
  COMM(help_writesession, writesession_command);
  COMM(help_zap,          zap_command);
  
  COMM(0, 0);              /* end mark for loops */
  
  if(pos != i + 1) {
    fprintf(stderr, "Wrong number of COMM's in init_comms_table()");
    exit(1);
  }
  
#undef COMM

};


/**************************************************************************/
/* parse input, check for TINTIN commands and aliases and send to session */
/**************************************************************************/
session *parse_input(const unsigned char *input, session *ses)
{
  unsigned char command[BUFFER_SIZE], arg[BUFFER_SIZE];
  unsigned char buf[BUFFER_SIZE], newcommand[BUFFER_SIZE];
  const unsigned char *input2, *commstr, *alias_res = 0;
  int comm_no, found, i, repeatif_times = 0;
  session *sesptr;
  window *win;
  
  if(*input == '\0') {
    if(ses->get_socket() >= 0)
      ses = ses->write_net((const unsigned char *)"");
    else
      ses = write_com_arg_mud((const uchar *)"", (const uchar *)"", ses);
    return ses; 
  }
  if((master_table.verbatimmode && *input != verbatim_char) ||
     (!master_table.verbatimmode && *input == verbatim_char)) {
    if(ses->get_socket() >= 0) {
      if(*input == verbatim_char)
	input++;
      ses = ses->write_net(input);
      return(ses);
    } else {
      sprintf((char *)buf, "#NO SESSION ACTIVE. USE THE %cSESSION-COMMAND TO START ONE.\n\r", tintin_char);
      ses->print(buf);
      if(session::get_sessions())
	ses->print((const uchar *)"#OR MAKE ONE OLDER SESSION ACTIVE ;)\n\r");
      return(ses);
    }
  }
  if(*input == verbatim_char)
    input++;

  input2 = input;
  while(*input2) {
    input2 = get_arg_stop_spaces(input2, buf);
    substitute_vars(buf, command, sizeof(command), FALSE, ses, FALSE);
    input2 = space_out(input2);
    
    if((!IS_SET(ses->ignore_flags, IGNORE_ALIAS)) &&
       (alias_res = check_all_aliases(command, input2, newcommand, ses))) {
      
      input2 = alias_res;
      
      if(master_table.aliascalls > MAX_ALIAS_CALLS) {
        ses->print((const uchar *)"#ERROR - Too many inside alias.(probably forever loop)\n\r");
	master_table.output_window = 0;
        break;
      } else
        master_table.aliascalls++;
      
      if(IS_SET(ses->echo_flags, ECHO_ALIAS)) {
	sprintf((char *)buf, "[alias: %s]\n\r", newcommand);
	ses->print(buf);
      }
      
      if(!(ses = parse_input(newcommand, ses))) {
        master_table.aliascalls--;
	master_table.output_window = 0;
        break;
      }
      
      master_table.aliascalls--;
      master_table.output_window = 0;
      
    } else {

      input2 = get_arg_all(input2, buf, ses);
      substitute_vars(buf, arg, sizeof(arg), FALSE, ses, FALSE);
      
      if(*command == tintin_char) {
	unsigned char tmp[BUFFER_SIZE];
	
	commstr = command + 1;
	
	if(!*commstr) {
	  ses->print((const uchar *)"#Please write command name too.\n\r");
	} else {
	  found = FALSE;
	  for(sesptr = master_table.main_session;
	      sesptr && !found;
	      sesptr = sesptr->get_next()) {
	    if(strcmp((const char *)sesptr->get_name(), (const char *)commstr) == 0) {
	      if(*arg) {   /* was: #sessioname commands */
		get_arg_with_spaces(arg, tmp);
		if(ses != sesptr)
		  parse_input(tmp, sesptr);
		else
		  ses = parse_input(tmp, ses);
	      } else {
		sprintf((char *)buf, "#SESSION '%s' ACTIVATED.\n\r", sesptr->get_name());
		ses->print(buf);
		
		if(ses->get_input() != sesptr->get_input()) {
		  window::set_active_win(sesptr->get_input());
		  sesptr->get_input()->move(sesptr->get_input()->get_cursor());
		}
		sesptr->get_input()->set_active_ses(sesptr);
		ses = sesptr;
	      }
	      found = TRUE;
	      break;
	    }
	  }
	  
	  win = find_window(commstr);
	  if(win) {
	    if(win != NULL_WIN) {
	      if(win != window::get_active_win()) {
		sesptr = (win->get_active_ses() ? win->get_active_ses() :
			  master_table.main_session);
		if(sesptr != window::get_active_win()->get_active_ses()) {
		  sprintf((char *)buf, "#SESSION '%s' ACTIVATED.\n\r", sesptr->get_name());
		  ses->print(buf);
		}
		ses->get_input()->next_window_active(win);
		window::set_active_win(win);
	      }
	      ses = window::get_active_win()->get_active_ses();
	    }
	    found = TRUE;
	    break;
	  }
	  
	  if(!found) {
	    if(isdigit(*commstr)) {
	      i = atoi((const char *)commstr);
	      if(i > 0) {
		get_arg_with_spaces(arg, tmp);
		while(i-- > 0)
		  if(!(ses = parse_input(tmp, ses)))
		    break;
	      } else {
		ses->print((const uchar *)"#YEAH RIGHT! GO REPEAT THAT YOURSELF DUDE.\n\r");
	      }
	    } else {
	      for(comm_no = 0; *commands[comm_no].name != '\n'; comm_no++) {
		if(is_abrev(commands[comm_no].name, commstr)) {
		  if(commands[comm_no].func) {
		    if(commands[comm_no].func == comm_repeatif) {
		      unsigned char condition[BUFFER_SIZE];
		      
		      get_arg_with_spaces(arg, condition);
		      if(!*condition) {
			ses->print((const uchar *)"#Error - Usage: #repeatif <CONDITION>\n\r");
		      } else {
			if(string_to_number(condition, ses)) {
			  if(repeatif_times > MAX_REPEATIF_CALLS) {
			    ses->print((const uchar *)"#ERROR - Too many repeatif times.(probably forever loop)\n\r"
				       "#Skipping to next command...\n\r");
			  } else {
			    repeatif_times++;
			    input2 = input;  // back to line begin :)
			  }
			}
		      }
		    } else {
		      ses = (commands[comm_no].func)(arg, ses);
		    }
		  } else {
		    ses->print((const uchar *)"#This command isn't installed yet.\n\r");
		  }
		  break;
		}
	      }
	      if(*commands[comm_no].name == '\n') {
		sprintf((char *)tmp, "#UNKNOWN TINTIN-COMMAND. (%c%s)\n\r",
			tintin_char, commstr);
		ses->print(tmp);
	      }
	    }
	  }
	}
	
      } else if(speedwalk && !*arg && is_speedwalk_dirs(command)) {
	ses = do_speedwalk(command, ses);
	
      } else {
	get_arg_with_spaces(arg, arg);
	ses = write_com_arg_mud(command, arg, ses);
      }
      
      master_table.output_window = 0;
      if(!ses) {
	break;
      }
    }
    if(*input2 == tintin_retchar)
      input2++;
  }

  return(ses);
}

/**********************************************************************/
/* return TRUE if commands only consists of capital letters N,S,E ... */
/**********************************************************************/
int is_speedwalk_dirs(const unsigned char *cp)
{
  int flag = FALSE;

  while(*cp) {
    if(*cp != 'n' && *cp != 'e' && *cp != 's' &&
       *cp != 'w' && *cp != 'u' && *cp != 'd' && !isdigit(*cp))
      return FALSE;
    if(!flag && !isdigit(*cp))
      flag = TRUE;
    cp++;
  }
  return(flag);
}

/**************************/
/* do the speedwalk thing */
/**************************/
session *do_speedwalk(const unsigned char *cp, session *ses)
{
  unsigned char sc[2] = {'x', '\0'}, *tmp, tmpbuf[BUFFER_SIZE];
  int i, times;
  
  while(*cp && ses) {
    if(isdigit(*cp)) {
      tmp = tmpbuf;
      while(*cp && isdigit(*cp))
        *tmp++ = *cp++;
      *tmp = '\0';
      if(*cp) {
        sc[0] = *cp;
        times = atoi((const char *)tmpbuf);
        cp++;
      } else {
        sc[0] = 'x';
        times = 0;
      } 
    } else {
      sc[0] = *cp;
      cp++;
      times = 1;
    }
    for(i = 0; i < times && ses; i++)
      ses = write_com_arg_mud(sc, (const uchar *)"", ses);
  }
  return(ses);
}

/**************************************************/
/* get all arguments - don't remove (s, )s and \s */
/**************************************************/
const unsigned char *get_arg_all(const unsigned char *s, unsigned char *arg,
				 session *ses)
{
  int inside = 0;
  int redirected = FALSE;
  
  s = space_out(s);
  
  master_table.output_window = 0;
  
  while(*s) {
    if(*s == tintin_retchar && !inside)
      break;
    
    if(!inside && *s == '>' && *(s + 1) == '>' && *(s + 2) == '>') {
      uchar winname[BUFFER_SIZE];
      s += 3;
      s = get_arg_stop_spaces(s, winname);
      
      if(redirected)
	ses->print((const uchar *)"#More than just one output redirection.\n\r");
      
      redirected = TRUE;
      
      if(!(master_table.output_window = find_window(winname)))
	ses->print((const uchar *)"#No such named window using default.\n\r");
    } else {
      
      if(*s == text_start) {
	*arg++ = *s++;
	inside++;
      } else if(*s == text_end) {
	*arg++ = *s++;
	if(inside)
	  inside--;
      } else if(*s == '\\') {
	*arg++ = *s++;
	if(*s)
	  *arg++ = *s++;
      } else {
	*arg++ = *s++;
      }
    }
  }
  *arg = '\0';
  return s;
}

/**************************************/
/* get all arguments - remove []s etc. */
/* Example:                           */
// In:  [this is it] way way \[hmmm;
// Out: this is it
// In:  this is it way way \[hmm;
// Out: this is it way way [hmm
/**************************************/
const unsigned char *get_arg_with_spaces(const unsigned char *s,
					 unsigned char *arg)
{
  int inside = 0;

  s = space_out(s);
  while(*s) {

    if(*s == tintin_retchar && !inside)
      break;

    if(*s == text_start) {
      if(inside)
        *arg++ = *s++;
      else
        s++;
      inside++;
    } else if(*s == text_end) {
      if(inside) {
        inside--;
	
	if(!inside)
	  break;
      }
      
      if(inside)
        *arg++ = *s++;
      else
        s++;
    } else if(*s == '\\') {
      if(inside)
        *arg++ = *s++;
      else
        s++;
      
      if(*s)
	*arg++ = *s++;
    } else {
      *arg++ = *s++;
    }
  }
  *arg = '\0';
  return s;
}

/*******************************************************/
/* get one argument - stop at spaces - remove ()s etc. */
/* example:                                            */
/* in: (yo yo) wayout and no more;                     */
/* out: yo yo                                          */
/*******************************************************/
const unsigned char *get_arg_stop_spaces(const unsigned char *s,
					 unsigned char *arg)
{
  int inside = 0;

  s = space_out(s);

  while(*s) {
    if((*s == tintin_retchar || isspace(*s)) && !inside)
      break;

    if(*s == text_start) {
      if(inside)
        *arg++ = *s++;
      else
        s++;
      inside++;
    } else if(*s == text_end) {
      if(inside)
        inside--;
      if(inside)
        *arg++ = *s++;
      else {
        s++;
      }
    } else if(*s == '\\') {
      if(inside)
        *arg++ = *s++;
      else
        s++;

      if(*s)
	*arg++ = *s++;
    } else {
      *arg++ = *s++;
    }
  }
  *arg = '\0';
  return s;
}

/*********************************************/
/* spaceout - advance ptr to next none-space */
/* return: ptr to the first none-space       */
/*********************************************/ 
const unsigned char *space_out(const unsigned char *s)
{
  while(isspace(*s))
    s++;
  return s;
}

/************************************/
/* send command+argument to the mud */
/************************************/
session *write_com_arg_mud(const uchar *command, const uchar *argument,
			   session *ses)
{
  unsigned char outtext[BUFFER_SIZE];
  int pos;
  
  if(ses->get_socket() < 0) {
    unsigned char buf[100];
    sprintf((char *)buf, "#NO SESSION ACTIVE. USE THE %cSESSION-COMMAND TO START ONE.\n\r", tintin_char);
    ses->print(buf);
    if(session::get_sessions())
      ses->print((const uchar *)"#OR MAKE ONE OLDER SESSION ACTIVE ;)\n\r");
  } else {
    check_insert_path(command, ses);
    pos = str_cat(outtext, 0, sizeof(outtext), command);
    if(*argument) {
      pos = str_cat(outtext, pos, sizeof(outtext), (const uchar *)" ");
      pos = str_cat(outtext, pos, sizeof(outtext), argument);
    }
    if(!(ses = ses->write_net(outtext)))
      return(0);
    
    outtext[pos] = '\n';
    if(ses->logfile)
      fwrite(outtext, pos + 1, 1, ses->logfile);
  }
  return(ses);
}

unsigned char *get_number(unsigned char *str, struct mathnode *queen,
			  int pos, int pri, int opp, session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp, *startstr, *tmpstr;
  int inside = 0;
  struct listnode **myvars, *ln;
  
  queen[pos].pri = pri;
  queen[pos].opp = opp;

  while(isspace(*str))
    str++;

  tmp = buf;
  if(*str == '(') {
    str++;
    inside = 0;
    while(*str && (*str != ')' || inside)) {
      if(*str == '(')
        inside++;
      else if(*str == ')')
        inside--;

      *tmp++ = *str++;
    }
    *tmp = '\0';
    if(*str == ')')
      str++;
    else {
      ses->print((const uchar *)"#Error - Unmatched parentheses error.\n\r");
    }

    queen[pos].number = string_to_number(buf, ses);
    queen[pos].var = 0;
    return(str);
  }

  startstr = str;

  while(*str && !isdigit(*str) && *str != '$')
    str++;

  tmpstr = str - 1;
  
  if(*str == '$') {
    while(*str == '$')
      str++;
    
    if(isdigit(*str)) {
      if(*vars[*str - '0'])
        queen[pos].number = string_to_number(vars[*str - '0'], ses);
      else
        queen[pos].number = 0;
      
      queen[pos].var = 0;
      str++;
    } else {
      while(*str != '$' && *str)
        *tmp++ = *str++;
      *tmp = '\0';
      
      if(*str)
        str++;
      
      myvars = &ses->variables;
      if((ln = searchnode_list(*myvars, buf)) != NULL) {
        queen[pos].number = string_to_number(ln->right, ses);
        queen[pos].var = ln;
      } else {
        queen[pos].number = 0;
        queen[pos].var = new listnode(myvars, buf, (unsigned char *)"0",
				      TYPE_VAR, 5);
	if(IS_SET(ses->message_flags, MESSAGE_VAR))
	  ses->print((const uchar *)"#NEW VARIABLE DEFINED.\n\r");
      }
    }
  } else {
    while(*str && isdigit(*str))
      *tmp++ = *str++;
    
    *tmp = '\0';
    
    if(*buf)
      queen[pos].number = atoi((const char *)buf);
    else
      queen[pos].number = 0;
    
    queen[pos].var = 0;
  }

  for(; tmpstr >= startstr; tmpstr--) {
    switch(*tmpstr) {
    case '!': {
      queen[pos].number = !queen[pos].number;
    } break;
    case '~': {
      queen[pos].number = ~queen[pos].number;
    } break;
    case '-': {
      queen[pos].number = -queen[pos].number;
    } break;
    case '+': {
    } break;
    case ' ': {
    } break;
    default: {
      sprintf((char *)buf, "#Error - Unknown char (%c) before number.\n\r", *tmpstr);
      ses->print(buf);
    } break;
    }
  }

  return(str);
}

unsigned char *pass_number(unsigned char *str, session *ses)
{
  int inside = 0;

  while(isspace(*str))
    str++;

  while(*str && !isdigit(*str) && *str != '$' && *str != '(')
    str++;

  if(*str == '(') {
    str++;
    inside = 0;
    while(*str && (*str != ')' || inside)) {
      if(*str == '(')
        inside++;
      else if(*str == ')')
        inside--;

      str++;
    }
    if(*str == ')')
      str++;
    else {
      ses->print((const uchar *)"#Error - Unmatched parentheses error.\n\r");
    }
    return(str);
  }

  if(*str == '$') {
    while(*str == '$')
      str++;

    while(*str != '$' && *str)
      str++;

    if(*str)
      str++;
  } else {
    while(*str && isdigit(*str))
      str++;
  }

  return(str);
}

int calculate_queen(struct mathnode *queen, int epos, session *ses)
{
  int i, j, prepos = 0, nextpos;
  struct mathnode *prenode;

  for(i = 1; i < 9; i++) {  /* no '||' or '&&' or '=' */
    for(prepos = 0, j = queen[0].next; j < epos; j = nextpos) {
      nextpos = queen[j].next;

      if(i == queen[j].pri) {
        prenode = &queen[prepos];

        switch(queen[j].opp) {
        case 1: {
          prenode->number *= queen[j].number;
        } break;
        case 2: {
          if(queen[j].number)
            prenode->number /= queen[j].number;
          else {
            ses->print((const uchar *)"#ERROR - division with 0.\n\r");
            queen[0].next = epos;
            queen[epos].prev = 0;
            queen[0].number = 0;
            return(0);
          }
        } break;
        case 3: {
          if(queen[j].number)
            prenode->number %= queen[j].number;
          else {
            ses->print((const uchar *)"#ERROR - division with 0.\n\r");
            queen[0].next = epos;
            queen[epos].prev = 0;
            queen[0].number = 0;
            return(0);
          }
        } break;
        case 4: {
          prenode->number += queen[j].number;
        } break;
        case 5: {
          prenode->number <<= queen[j].number;
        } break;
        case 6: {
          prenode->number >>= queen[j].number;
        } break;
        case 7: {
          prenode->number = (prenode->number < queen[j].number);
        } break;
        case 8: {
          prenode->number = (prenode->number <= queen[j].number);
        } break;
        case 9: {
          prenode->number = (prenode->number > queen[j].number);
        } break;
        case 10: {
          prenode->number = (prenode->number >= queen[j].number);
        } break;
        case 11: {
          prenode->number = (prenode->number == queen[j].number);
        } break;
        case 12: {
          prenode->number = (prenode->number != queen[j].number);
        } break;
        case 13: {
          prenode->number &= queen[j].number;
        } break;
        case 14: {
          prenode->number ^= queen[j].number;
        } break;
        case 15: {
          prenode->number |= queen[j].number;
        } break;
        default: {
          printf("BUG BUG BUG.........*slap* Caster\n");
          exit(1);
        } break;
        } /* switch */

        queen[j].pri = 0;
        prenode->next = queen[j].next;
        queen[prenode->next].prev = prepos;

      } else if(queen[j].pri == 0) {
        printf("BUG BUG BUG.........*slap* Caster (pri = 0)\n");
        exit(1);
      } else
        prepos = j;
    }
  }
  return(prepos);
}

int string_to_number(unsigned char *str, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  int stop, prepos, nextpos, i, j;
  int queenpos = 0;
  struct mathnode queen[101], *prenode;

  static int times = 0;

  if(++times > 6) {
    ses->print((const uchar *)"#ERROR Too many inside string_to_number() calls.\n\r"
	       "#variables inside variable ?!?\n\r");
    return(0);
  }

  for(i = 1;i < 100; i++) {
    queen[i].next = i + 1;
    queen[i].prev = i - 1;
  }
  queen[0].next = 1;
  queen[100].prev = 99;

  queenpos = 0;
  str = get_number(str, queen, queenpos++, 0, 0, ses);

  stop = FALSE;
  while(*str) {
    while(isspace(*str))
      str++;

    switch(*str++) {
    case '*': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 20, ses);
      } else
        str = get_number(str, queen, queenpos++, 1, 1, ses);
    } break;
    case '/': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 21, ses);
      } else
        str = get_number(str, queen, queenpos++, 1, 2, ses);
    } break;
    case '%': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 22, ses);
      } else
        str = get_number(str, queen, queenpos++, 1, 3, ses);
    } break;
    case '+': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 23, ses);
      } else
        str = get_number(str, queen, queenpos++, 2, 4, ses);
    } break;
    case '-': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 24, ses);
      } else {
        str = get_number(str, queen, queenpos, 2, 4, ses);
        queen[queenpos].number = -queen[queenpos].number;
        queenpos++;
      }
    } break;
    case '<': {
      switch(*str) {
      case '<': {
        str++;
        if(*str == '=') {
          str++;
          str = get_number(str, queen, queenpos++, 12, 25, ses);
        } else
          str = get_number(str, queen, queenpos++, 3, 5, ses);
      } break;
      case '=': {
        str++;
        str = get_number(str, queen, queenpos++, 4, 8, ses);
      } break;
      default: {
        str = get_number(str, queen, queenpos++, 4, 7, ses);
      } break;
      } /* switch */
    } break;
    case '>': {
      switch(*str) {
      case '>': {
        str++;
        if(*str == '=') {
          str++;
          str = get_number(str, queen, queenpos++, 12, 26, ses);
        } else
          str = get_number(str, queen, queenpos++, 3, 6, ses);
      } break;
      case '=': {
        str++;
        str = get_number(str, queen, queenpos++, 4, 10, ses);
      } break;
      default: {
        str = get_number(str, queen, queenpos++, 4, 9, ses);
      } break;
      } /* switch */
    } break;
    case '=': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 5, 11, ses);
      } else {
        str = get_number(str, queen, queenpos++, 12, 19, ses);
      }
    } break;
    case '!': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 5, 12, ses);
      } else {
        ses->print((const uchar *)"#Error - Wrong position for '!' operation.\n\r");
	times--;
        return(0);
      }
    } break;
    case '&': {
      switch(*str) {
      case '&': {
        str++;
        prepos = calculate_queen(queen, queenpos, ses);
        if(queen[prepos].number)
          str = get_number(str, queen, queenpos++, 9, 16, ses);
        else {
          str = pass_number(str, ses);
          while(*str) {
            while(isspace(*str))
              str++;

            if(*str == '|' && *(str + 1) == '|')
              break;
            else
              str = pass_number(str, ses);
          }
        }
      } break;
      case '=': {
        str++;
        str = get_number(str, queen, queenpos++, 12, 27, ses);
      } break;
      default: {
        str = get_number(str, queen, queenpos++, 6, 13, ses);
      } break;
      } /* switch */
    } break;
    case '^': {
      if(*str == '=') {
        str++;
        str = get_number(str, queen, queenpos++, 12, 28, ses);
      } else
        str = get_number(str, queen, queenpos++, 7, 14, ses);
    } break;
    case '|': {
      switch(*str) {
      case '|': {
        str++;
        prepos = calculate_queen(queen, queenpos, ses);
        if(queen[prepos].number) {
          stop = TRUE;
        } else
          str = get_number(str, queen, queenpos++, 10, 17, ses);
      } break;
      case '=': {
        str++;
        str = get_number(str, queen, queenpos++, 12, 29, ses);
      } break;
      default: {
        str = get_number(str, queen, queenpos++, 8, 15, ses);
      } break;
      } /* switch */
    } break;
    case ',': {
      str = get_number(str, queen, queenpos++, 13, 30, ses);
    } break;
    default: {
      stop = TRUE;
    } break;
    } /* switch */
    if(stop)
      break;
  }
  prepos = calculate_queen(queen, queenpos, ses);
  if(prepos != 0) {
    for(i = 9; i < 11; i++) {
      for(prepos = 0, j = queen[0].next; j < queenpos; j = nextpos) {
        nextpos = queen[j].next;

        if(i == queen[j].pri) {
          prenode = &queen[prepos];

          switch(queen[j].opp) {
          case 16: {
            prenode->number = (prenode->number && queen[j].number);
          } break;
          case 17: {
            prenode->number = (prenode->number || queen[j].number);
          } break;
          default: {
            printf("BUG BUG BUG.........*slap* Caster\n");
            exit(1);
          } break;
          } /* switch */

          queen[j].pri = 0;
          prenode->next = queen[j].next;
          queen[prenode->next].prev = prepos;
        } else if(queen[j].pri == 0) {
          printf("BUG BUG BUG (|| or &&) .....*slap* Caster (pri = 0)\n");
          exit(1);
        } else
          prepos = j;
      }
    }
    for(i = queen[queenpos].prev; i > 0; i = nextpos) {
      nextpos = queen[i].prev;

      if(queen[i].pri == 12) {
        prenode = &queen[queen[i].prev];
        if(prenode->var) {
          switch(queen[i].opp) {
          case 19: {
            prenode->number = queen[i].number;
          } break;
          case 20: {
            prenode->number *= queen[i].number;
          } break;
          case 21: {
            prenode->number /= queen[i].number;
          } break;
          case 22: {
            prenode->number %= queen[i].number;
          } break;
          case 23: {
            prenode->number += queen[i].number;
          } break;
          case 24: {
            prenode->number -= queen[i].number;
          } break;
          case 25: {
            prenode->number <<= queen[i].number;
          } break;
          case 26: {
            prenode->number >>= queen[i].number;
          } break;
          case 27: {
            prenode->number &= queen[i].number;
          } break;
          case 28: {
            prenode->number ^= queen[i].number;
          } break;
          case 29: {
            prenode->number |= queen[i].number;
          } break;
          default: {
            printf("BUG BUG BUG.........*slap* Caster\n");
            exit(1);
          } break;
          } /* switch */

          if(prenode->var->right)
            free(prenode->var->right);
          sprintf((char *)buf, "%d", prenode->number);
          prenode->var->right = (uchar *)strdup((const char *)buf);
        } else {
          ses->print((const uchar *)"#ERROR - there isn't any variable were to place value\n\r");
        }
        queen[i].pri = 0;
        prenode->next = queen[i].next;
        queen[prenode->next].prev = queen[i].prev;
      }
    }

    for(i = queen[0].next; i < queenpos; i = nextpos) {
      nextpos = queen[i].next;

      prenode = &queen[queen[i].prev];

      if(queen[i].opp == 30) { /* , */
        prenode->number = queen[i].number;
      } else {
        printf("BUG BUG BUG.........*slap* Caster\n");
        exit(1);
      }
      queen[i].pri = 0;
      prenode->next = queen[i].next;
      queen[prenode->next].prev = queen[i].prev;
    }
  }
  
  times--;
  return(queen[0].number);
}

/* ******************************************************************** */

unsigned char *own_asctime(void)
{
  static unsigned char timebuf[10];
  struct tm *timedata;
  time_t timep;
  
  timep = time(0);
  timedata = localtime(&timep);
  
  sprintf((char *)timebuf, "%2.2d:%2.2d:%2.2d",
	  timedata->tm_hour,
	  timedata->tm_min,
	  timedata->tm_sec);
  
  return(timebuf);
}
