#include "color.h"

void redraw_screen(void);
void prompt(void);
void get_screen_size(void);

void cbreak_noecho(void);
void nocbreak_echo(void);
void init_termcap();

session *comm_split(const unsigned char *arg, session *ses);
session *comm_output(const unsigned char *arg, session *ses);
session *comm_input(const unsigned char *arg, session *ses);
session *comm_close(const unsigned char *arg, session *ses);
session *comm_wininfo(const unsigned char *arg, session *ses);
session *comm_winsize(const unsigned char *arg, session *ses);

window *find_window(const unsigned char *name);

struct position {
  int row, col;
};

class window {
public:
  window();
  window(const unsigned char *nam, const int size, window *win);
  ~window();

  void move(const position *pos);
  void move(const int row, const int col);
  void moveforward();
  void movebackward();
  void scroll_region(const scrollsize *screg);
  void scroll_region(const int start, const int end);
  void clear();
  void clearend();
  void clearstart();
  void redraw(const int position = 0); // cursor 0 = center, -1 up and 1 = down
  void bell();
  
  void print(const signed char *text) {
    print((const unsigned char *)text);
  }
  void print(const unsigned char *text);
  void print(const unsigned char *text, const unsigned char *add);
  
  int change_time(const int delay);
  void print_time(const int all, const int bell = FALSE);
  void print_win_hist_line(const int all, unsigned char *text = 0);
  
  void print_prompt(const int all, const int log = TRUE);
  void remove_user_text();
  
  const int get_hist_pos() const      { return window_hist_pos; }
  void add_hist_pos()                 { window_hist_pos++; }
  void sub_hist_pos()                 { window_hist_pos--; }
  void set_hist_pos(const int n)      { window_hist_pos = n; }
  const session *get_hist_on() const  { return window_hist_on; }
  const int get_hist_clip() const     { return window_hist_clip; }

  void hist_movebackward();
  void hist_moveforward();
  void hist_backward(int amount = 1);
  void hist_forward(int amount = 1);
  void hist_home();
  void hist_end();
  void hist_pageup();
  void hist_pagedown();
  int hist_search(const unsigned char *str, const int forward = 1);
  void hist_set_mark();
  void hist_copy_clip();
  void hist_rem_clip_color();
  int start_win_history(session *ses);
  void exit_win_history();
  void max_size();
  int change_size(const int newsize);
  void next_window_active(window *newact = 0);
  
  const unsigned char *get_name() const   { return name; }
  
  const scrollsize *get_size() const      { return &WinSize; }
  const scrollsize *get_scroll() const    { return &Scroll; }
  const position *get_cursor() const      { return &cursor; }
  const int get_ptexrem() const           { return ptexrem; }
  void set_ptexrem()                      { ptexrem = TRUE; }
  void rem_ptexrem()                      { ptexrem = FALSE; }
  
  void set_active_ses(session *ses)       { active_ses = ses; }
  session *get_active_ses() const         { return active_ses; }
  
  static window *get_windows()            { return windows; }
  static void set_active_win(window *win) { active = win; }
  const window *get_lastactive() const    { return lastactive; }
  static window *get_active_win()         { return active; }
  static const int get_cols()             { return ScrCols; }
  static const int get_rows()             { return ScrRows; }
  static const position *get_scrcursor()  { return &Scrcursor; }
  static void get_screen_size();
  
  window *get_next() const        { return next; }

  time_t lastupdate;
  const int get_delay() const     { return timedelay; }
  const int get_time_on() const   { return(timedelay >= 0 ? TRUE : FALSE); }

  static const unsigned char *get_scrcolormode() {
    return window::Scrcolormode;
  }
  const unsigned char *get_colormode() { return colormode; }
  void check_color();

private:
  unsigned char *name;
  unsigned char colormode[4];
  
  scrollsize WinSize;
  scrollsize Scroll;
  position cursor;
  int timedelay;            // -1 = not active, 0 = no clock
  int ptexrem, autowarp;
  int window_hist_pos, window_hist_clip;
  session *window_hist_on;

  session *active_ses;
  
  static unsigned char Scrcolormode[4];
  static int ScrRows;
  static int ScrCols;
  static scrollsize ScrScroll;
  static position Scrcursor;
  
  static window *windows;

  window *lastactive;               // this is for print_time()
  static window *active;
  window *next;
};

inline void window::check_color()
{
  unsigned char color[BUFFER_SIZE];
  
  bits_to_color(window::Scrcolormode, colormode, color);
  if(*color) {
    printf("%s", color);
    window::Scrcolormode[0] = colormode[0];
    window::Scrcolormode[1] = colormode[1];
    window::Scrcolormode[2] = colormode[2];
    window::Scrcolormode[3] = colormode[3];
    fflush(stdout);
  }
}
