/*********************************************************************/
/* file: substitute.c - functions related to the substitute command  */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <string.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#endif
#include <ctype.h>

#include "structs.h"
#include "action.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "substitute.h"
#include "utility.h"
#include "utils.h"

extern unsigned char *letters;
extern int letters_number;

extern struct master_table_data master_table;
extern unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,....$9 variables */
extern unsigned char *colors[];

/***************************/
/* the #substitute command */
/***************************/
session *parse_sub(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE];
  unsigned char result[BUFFER_SIZE/2], pritext[BUFFER_SIZE], letter;
  int found, len, amount, pos, nums[10], numpos, strpos, ret_value, pri = 15;
  struct listnode **mysubs, *ln;

  mysubs = &ses->actions;

  arg=get_arg_stop_spaces(arg, left);
  arg=get_arg_stop_spaces(arg, right);
  arg=get_arg_with_spaces(arg, pritext);
  if(*pritext) {
    pri = string_to_number(pritext, ses);
    if(pri < 0 || pri > 20) {
      ses->print((const uchar *)"#Priority must be 0-20!! (using default value)\n\r");
      pri = 15;
    }
  }

  if(!*left) {
    ses->print((const uchar *)"#THESE SUBSTITUTES (and actions) HAS BEEN DEFINED:\n\r");
    show_list(*mysubs, ses);
  }
  
  else if(*left && !*right) {
    for(found = FALSE,ln = searchnode_list2(*mysubs, left); ln;
	ln = searchnode_list2(ln->next, left)) {
      if(ln->type == TYPE_SUBS || ln->type == TYPE_GAGS) {
	found = TRUE;
	ln->show(ses);
      }
    }
    if(!found)
      ses->print((const uchar *)"#THAT SUBSTITUTE IS NOT DEFINED.\n\r");
  }

  else {
    if((ln = searchnode_list(*mysubs, left)) != NULL) {
      if(ln->type == TYPE_SUBS || ln->type == TYPE_GAGS)
	delete ln;
      else if((ln = searchnode_list(ln->next, left)) != NULL)
	delete ln;
    }
    
    if(strcmp("<*GAG*>", (const char *)right) == 0)
      ln = new listnode(mysubs, left, right, TYPE_GAGS, pri);
    else
      ln = new listnode(mysubs, left, right, TYPE_SUBS, pri);

    if(IS_SET(ses->message_flags, MESSAGE_SUBS))
      ses->print((const uchar *)"#OK. SUBSTITUTE DEFINED.\n\r");
    found = FALSE;
    len = strlen((const char *)left);
    if(len < 2)
      found = TRUE;
    amount = search_vars(right, result);
    
    if(!found && *left != '^') {   // check that it will start ^ or variable
      if(*left != '$' || !isdigit(*(left + 1))) {
	found = TRUE;
      }
    }
    if(!found) {                   // check that it will end ^ or variable
      len = strlen((const char *)left);
      if(*(left + len - 1) != '^') {
	if(*(left + len - 2) != '$' || !isdigit(*(left + len - 1))) {
	  found = TRUE;
	}
      }
    }
    if(!found) {                   // check variables
      for(pos = 0; pos < 10; pos++)
	nums[pos] = FALSE;
      
      for(len = 0; *(left + len) && !found; len++) {
	if((ret_value = is_variable(left + len))) {
	  if(ret_value < 0)
	    letter = left[len + 1];
	  else
	    letter = left[len + 3];
	  numpos = (int)((int)letter - (int)'0');

	  if(nums[numpos]) {
	    found = TRUE;
	  } else {
	    nums[numpos] = TRUE;
	    
	    for(pos = 0; pos < amount; pos++)
	      if(letter == result[pos])
		break;
	    if(pos == amount)
	      found = TRUE;
	  }
	}
      }
    }
    if(found && ln->type == TYPE_SUBS) {
      ses->print((const uchar *)"#But You WILL LOSE text with this substitute...\n\r");
      ses->print((const uchar *)"#And that probably wasn't what you wanted.\n\r");
    }
    found = FALSE;
    strpos = 0;
    for(pos = 0; pos < amount; pos++) {
      numpos = (int)((int)result[pos] - (int)'0');
      if(!nums[numpos]) {
	if(!found) {
	  found = TRUE;
	  strpos = str_cat(left, 0, sizeof(left), (const uchar *)"#Following variables will be replaced with random text:\n\r# ");
	} else {
	  strpos = str_cat(left, strpos, sizeof(left), (const uchar *)", ");
	}
	sprintf((char *)right, "$%d", numpos);
	strpos = str_cat(left, strpos, sizeof(left), right);
      }
    }
    if(found) {
      strpos = str_cat(left, strpos, sizeof(left), (const uchar *)"\n\r");
      ses->print(left);
    }
  }
  return(ses);
}

/*****************************/
/* the #unsubstitute command */
/*****************************/
session *unsubstitute_command(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE];
  int found;
  struct listnode **mysubs, *ln, *next;

  mysubs = &ses->actions;

  arg = get_arg_with_spaces(arg, left);

  for(found = FALSE,ln = searchnode_list2(*mysubs, left); ln;
      ln = searchnode_list2(next, left)) {
    next = ln->next;
    if(ln->type == TYPE_SUBS || ln->type == TYPE_GAGS) {
      found = TRUE;
      delete ln;
      if(IS_SET(ses->message_flags, MESSAGE_SUBS))
	ses->print((const uchar *)"#OK. SUBSTITUTE DELETED\n\r");
    }
  }
  if(!found)
    ses->print((const uchar *)"#THAT SUBSTITUTE IS NOT DEFINED.\n\r");
  
  return(ses);
}
